--[[
=========================
CommentaryServer module original idea: nesa24
Game research and programming: juce, nesa24, predator002
Requires: sider.dll 7.0.2.0+
Requires: GameVersion 1.3.0
=========================
--]]

local m = { version = "v0.5" }

local patterns = {
    "sound\\awb\\00_TEAM.awb",
    "sound\\awb\\10_PLAYER.awb",
    "sound\\awb\\20_MAIN.awb",
    "sound\\awb\\30_MAIN.awb",
    "sound\\awb\\40_MAIN.awb",
    "sound\\awb\\50_MAIN.awb",
    "sound\\awb\\60_ADD.awb",
    "sound\\acb\\00_TEAM.acb",
    "sound\\acb\\10_PLAYER.acb",
    "sound\\acb\\20_MAIN.acb",
    "sound\\acb\\30_MAIN.acb",
    "sound\\acb\\40_MAIN.acb",
    "sound\\acb\\50_MAIN.acb",
    "sound\\acb\\60_ADD.acb",
    "sound\\config\\Rc\\rc_config.bin",
    "sound\\config\\Control\\ng_list.xml",
    "sound\\config\\Control\\tree_control.xml",
}

local language_codes = {
    ["eng"] = 0,      -- English
    ["fra"] = 1,      -- French
    ["ger"] = 2,      -- German
    ["ita"] = 3,      -- Italian
    ["spa"] = 4,      -- Spanish
    ["jpn"] = 5,      -- Japanese
    ["kor"] = 6,      -- Korean
    ["chi"] = 7,      -- Traditional Chinese
    ["use"] = 8,      -- US English
    ["pol"] = 9,      -- Polish
    ["por"] = 0x0a,   -- Portuguese
    ["mex"] = 0x0b,   -- Mexican Spanish
    ["rus"] = 0x0c,   -- Russian
    ["gre"] = 0x0d,   -- Greek
    ["swe"] = 0x0e,   -- Swedish
    ["nld"] = 0x0f,   -- Dutch
    ["bra"] = 0x10,   -- Brazilian Portuguese
    ["tur"] = 0x11,   -- Turkish
    ["zha"] = 0x12,   -- Simplified Chinese
    ["ara"] = 0x13,   -- Arabic
    ["ags"] = 0x14,   -- Argentine Spanish
    ["chl"] = 0x15,   -- Chilean Spanish
    ["man"] = 0x16,   -- Mandarin Chinese
    ["can"] = 0x17,   -- Cantonese Chinese
}

local content_root
local curr_lang
local user_lang
local last_loaded
local structure_addr
local read_language_addr
local patched

local map_teams = {}
local map_competitions = {}

-- map priority: the order in which the maps are used to select language.
-- if teams_first is set to false, then: first look in map_competitions, then in map_teams
-- if teams_first is set to true, then: first look in map_teams, then in map_competitions
-- (if entry is not found in the first map, only then the second map is checked.)
local teams_first = false


local function trim(s)
    return s:gsub("^%s*(.-)%s*$", "%1")
end


local function split(s, inSplitPattern)
    local outResults = {}
    local theCommentStart = string.find( s, "#", 1 )
    local data = s
    if theCommentStart ~= nil then
        data = string.sub(s, 1, theCommentStart-1)
    end

    local theStart = 1
    local theSplitStart, theSplitEnd = string.find( data, inSplitPattern, theStart )
    while theSplitStart do
        outResults[#outResults+1] = trim(string.sub( data, theStart, theSplitStart-1 ))
        theStart = theSplitEnd + 1
        theSplitStart, theSplitEnd = string.find( data, inSplitPattern, theStart )
    end
    outResults[#outResults+1] = trim(string.sub( data, theStart ))
    return outResults
end


local function len(map)
    local count = 0
    if map ~= nil then
        for k, _ in pairs(map) do
            count = count + 1
        end
    end
    return count
end


local function load_map_txt(map, filename)
    local delim = ","
    local f = io.open(content_root .. "\\" .. filename)
    if not f then
        log("WARN: map file: " .. content_root .. "\\" .. filename .. " not found. Skipping")
        return
    end
    for line in f:lines() do
        local fields = split(line, delim)
        if #fields > 1 then
            local key = tonumber(fields[1])
            if key ~= nil and fields[2] ~= nil and fields[2] ~= "" then
                map[key] = fields[2]
            end
        end
    end
    f:close()
    log(string.format("%s: %d entries", filename, len(map)))
end


function m.get_filepath(ctx, filename, key)
    for i,pattern in ipairs(patterns) do
        if string.match(filename, pattern) then
            last_loaded = filename
            if not patched and pattern == "sound\\acb\\00_TEAM.acb" then
                user_lang = filename:sub(1,3)
                log(string.format("new user_lang = %s", user_lang))
            end
            log(string.format("****** using: %s", filename))
            if patched and pattern == "sound\\awb\\40_MAIN.awb" then
                m.clear_patch(ctx)
            end
            return
        end
    end
end


function m.set_teams(ctx, home, away)
    -- Look up language
    if teams_first then
        -- map_teams has priority
        curr_lang = map_teams[home]
        curr_lang = curr_lang or map_competitions[ctx.tournament_id]
    else
        -- map_competitions has priority
        curr_lang = map_competitions[ctx.tournament_id]
        curr_lang = curr_lang or map_teams[home]
    end

    log(string.format("user_lang = %s", user_lang))
    log(string.format("curr_lang = %s", curr_lang))

    -- force ACB reload
    local addr = memory.unpack("u64", memory.read(structure_addr, 8))
    memory.write(addr + 0x94, "\x08")

    -- enforce language
    local lang_code = language_codes[curr_lang or user_lang] or 0x18
    memory.write(read_language_addr, "\x48\xb9" .. memory.pack("u8", lang_code) .. "\x00\x00\x00\x00\x00\x00\x00")
    patched = true
end


function m.clear_patch(ctx)
    -- restore replaced code
    memory.write(read_language_addr, "\x48\x8b\x47\x70\x48\x8b\x00\x8b\x48\x18")
    log("replaced code for language reading restored")
    patched = false
end


function m.overlay_on(ctx)
    return string.format("%s | user default: %s, selected: %s, last loaded: %s",
        m.version, user_lang, curr_lang, last_loaded)
end


function m.init(ctx)
    content_root = ctx.sider_dir .. "content\\commentary-server\\"             -- set root folder
    load_map_txt(map_competitions, "map_competitions.txt")                     -- load competitions map
    load_map_txt(map_teams, "map_teams.txt")                                   -- load team map
    log(string.format("teams_first: %s", teams_first))

    ctx.register("set_teams", m.set_teams)                                     -- set check team map procedure
    ctx.register("livecpk_get_filepath", m.get_filepath)                       -- check file loads
    ctx.register("overlay_on", m.overlay_on)

    --[[
    0000000140C5672D | 45:0FB6F6                | movzx r14d,r14b                            |
    0000000140C56731 | 83F8 13                  | cmp eax,13                                 |
    0000000140C56734 | 45:0F44F4                | cmove r14d,r12d                            |
    0000000140C56738 | 48:8B0D 5198AA02         | mov rcx,qword ptr ds:[1436FFF90]           |
    --]]
    local pos = memory.search_process("\x45\x0f\xb6\xf6\x83\xf8\x13\x45\x0f\x44\xf4\x48\x8b\x0d")
    if not pos then
        error("unable to find data-structure address")
    end
    pos = pos + 14
    local offset = memory.unpack("i32", memory.read(pos, 4))
    structure_addr = pos + 4 + offset
    log(string.format("structure address: %s", memory.hex(structure_addr)))

    --[[
    0000000140EA1468 | 48:8B47 70                      | mov rax,qword ptr ds:[rdi+70]           | start reading new language
    0000000140EA146C | 48:8B00                         | mov rax,qword ptr ds:[rax]              |
    0000000140EA146F | 8B48 18                         | mov ecx,dword ptr ds:[rax+18]           |
    0000000140EA1472 | E8 B9560300                     | call pes2021.140ED6B30                  |
    --]]
    read_language_addr = memory.search_process("\x48\x8b\x47\x70\x48\x8b\x00\x8b\x48\x18\xe8")
    if not read_language_addr then
        error("unable to find code position for reading language")
    end
    log(string.format("read-language address: %s", memory.hex(read_language_addr)))
end

return m
